function msg = check_is( obj , obj_type , msg , varargin )
% Copyright (C) 2005,2006,2007 Daniele de Rigo 
% 
% This file is part of Mastrave.
%
% Mastrave is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License along
% with this program; if not, write to the Free Software Foundation, Inc.,
% 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
% 
% ---------------------------------------------------------------------------
% 
%  msg = check_is( obj , obj_type , msg , ... )
% 
% Checks if the variable <obj> belongs to the category <obj_type>.
% If yes, returns to the caller.
% If not, the behaviour of the function depends on the number of output
% arguments:
%    if nargout==0   throws an error showing the message which format is <msg>.
%    if nargout==1   returns the error message which format is <msg>.
% If other arguments are passed after <msg>, they are managed like the
% optional arguments in the function "sprintf" (see the help of "sprintf"
% for more details about the format of <msg> and the optional arguments).
% 
% List of the implemented categories for <obj_type>
% 
%   <obj_type>          constraint to verify 
% 
%   <model>             class( <obj> ) must be the same as class( <model> ) 
%                       and if <model> is a cell-array or a struct, each
%                       element of <model> must recursively have the same
%                       class of the corresponding element of <obj>.
%                       <model> cannot be a string: if you need to check if
%                       <obj> is a string, use instead something like
%                           check_is( <obj> , 'string' , 'must be a string!' )
%   'true'              synonym of:  logical(   <obj> )
%   'false'             synonym of:  ~logical(  <obj> )
%   'logical'           synonym of:  islogical( <obj> )
%   'numeric'           synonym of:  isnumeric( <obj> ) | islogical( <obj> )
%   'char' , 'string'   synonym of:  ischar(    <obj> )
%   'cell'              synonym of:  iscell(    <obj> )
%   'struct'            synonym of:  isstruct(  <obj> )
%   'int8'              synonym of:  isa( <obj> , 'int8'   )
%   'uint8'             synonym of:  isa( <obj> , 'uint8'  )
%   'int16'             synonym of:  isa( <obj> , 'int16'  )
%   'uint16'            synonym of:  isa( <obj> , 'uint16' )
%   'int32'             synonym of:  isa( <obj> , 'int32'  )
%   'uint32'            synonym of:  isa( <obj> , 'uint32' )
%   'cellstring'        <obj> must be a 'char' or a cell array which elements 
%                       <obj_type> must be 'char'
%   'cellnumeric'       <obj> must be 'numeric' or a cell array which elements 
%                       <obj_type> must be 'numeric'
%   'cellstring-r'      <obj> must be a 'char' or a cell array which elements 
%                       <obj_type> must be 'cellstring-r'
%   'cellnumeric-r'     <obj> must be 'numeric' or a cell array which elements 
%                       <obj_type> must be 'cellnumeric-r'
%   'positive'          <obj> must be numeric and each of its elements must be
%                       greater than zero
%   'nonpositive'       <obj> must be numeric and each of its elements must be
%                       less or equal to zero
%   'negative'          <obj> must be numeric and each of its elements must be
%                       less than zero
%   'nonnegative'       <obj> must be numeric and each of its elements must be
%                       greater or equal to zero
%   'integer'           <obj> must be numeric and each of its elements must be
%                       convertible to an integer without loss of information
%   'natural'           <obj> must be numeric and each of its elements must be
%                       convertible to a non-negative integer without loss of 
%                       information
%   'natural_nonzero'   <obj> must be numeric and each of its elements must be
%                       convertible to a positive integer without loss of 
%                       information
%   'scalar'            synonym of:  lenght( <obj> )==1 
% 
% version: 0.7

if nargin < 3   
   fprintf(2,  'Error: too few arguments\n'  );
   fprintf(2,  'Usage: check_is( obj , obj_type , msg , ... )\n'  );
   error(  ' '  );
end

if ~ischar(obj_type)
%   fprintf(2,  'Error: the 2nd argument must be a string\n'  );
%   fprintf(2,  'Usage: check_is( obj , obj_type , msg , ... )\n'  );
%   error(  ' '  );
   model    =  obj_type ;
   obj_type =  'model'  ;
end

if ~ischar(msg)
   fprintf(2,  'Error: the 3rd argument must be a string\n'  );
   fprintf(2,  'Usage: check_is( obj , obj_type , msg , ... )\n'  );
   error(  ' '  );
end

err = 0;
obj_type = obj_type(:)';
switch obj_type 
case  'model'  
   [ answer, bad_elem ] = has_model( obj , model , inputname(1) );
   err = ~answer;
   if length(bad_elem)
      msg = sprintf(  '%s\nat position: %s'  , msg , bad_elem );
   end
case  'true'  
   err = ~is_numeric(obj);  
   if ~err
      err = any(~obj);
   end
case  'false'  
   err = ~is_numeric(obj);  
   if ~err
      err = any(obj);
   end
case  'logical'  
   err = ~islogical(obj);
case  'numeric'  
   err = ~is_numeric(obj);   
case  {  'char'  ,  'string'  }  
   err = ~ischar(obj);
case  {  'cell'   ,  'struct'  ,...
         'int8'   ,  'uint8'   ,...
         'int16'  ,  'uint16'  ,...
         'int32'  ,  'uint32'  }  
   err = ~isa( obj , obj_type );
case  'cellstring'  
   [ answer, bad_elem ] = is_cell_of( obj ,  'char'    , inputname(1) );
   err = ~answer;
   if length(bad_elem)
      msg = sprintf(  '%s\nat position: %s'  , msg , bad_elem );
   end
case  'cellnumeric'  
   [ answer, bad_elem ] = is_cell_of( obj ,  'double'  , inputname(1) );
   err = ~answer;
   if length(bad_elem)
      msg = sprintf(  '%s\nat position: %s'  , msg , bad_elem );
   end
case  'cellstring-r'  
   [ answer, bad_elem ] = is_recursive_cell_of( obj ,  'char'    , inputname(1) );
   err = ~answer;
   if length(bad_elem)
      msg = sprintf(  '%s\nat position: %s'  , msg , bad_elem );
   end
case  'cellnumeric-r'  
   [ answer, bad_elem ] = is_recursive_cell_of( obj ,  'double'  , inputname(1) );
   err = ~answer;
   if length(bad_elem)
      msg = sprintf(  '%s\nat position: %s'  , msg , bad_elem );
   end
case  'positive'  
   err = ~is_numeric(obj);
   if ~err
      err = any(obj<=0);
   end
case  'nonpositive'  
   err = ~is_numeric(obj);
   if ~err
      err = any(obj>0);
   end
case  'negative'  
   err = ~is_numeric(obj);
   if ~err
      err = any(obj>=0);
   end
case  'nonnegative'  
   err = ~is_numeric(obj);
   if ~err
      err = any(obj<0);
   end
case  'integer'  
   err = ~is_numeric(obj);
   if ~err
      err = any(obj~=floor(obj));
   end
case  'natural'  
   err = ~is_numeric(obj);
   if ~err
      err = any( obj~=floor(obj) | any(obj<0) );
   end
case  'natural_nonzero'  
   err = ~is_numeric(obj);
   if ~err
      err = any( obj~=floor(obj) | any(obj<=0) );
   end
case  'scalar'  
   err = lenght(obj)~=1;
otherwise
   fprintf(2,  'Error: check for type <%s> not implemented\n'  ,obj_type);
   error(  ' '  );
end


if err
   msg = sprintf([  msg  '\n'  ], varargin{:} );
else
   msg = [];
end

if ~nargout
   error(msg);
end



function [ answer, msg ] = has_model( obj , model , obj_name ) 
   msg    =  ''  ;
   answer = strcmp( class(obj) , class(model) );
   if answer
      switch class(model)
      case  'struct'  
         %    [sorted, idx] = sort( my_string_cell ) is not portable.
         % Use instead:
         %    [sorted, idx] = sortrows( char( my_string_cell ) );
         %     sorted       = cellstr( sorted );
         [obj_fn, obj_id] = sortrows( char(fieldnames(  obj  )) );
         [mod_fn, mod_id] = sortrows( char(fieldnames( model )) );
         obj_fn   = cellstr( obj_fn );
         mod_fn   = cellstr( mod_fn );
         answer   = isequal( obj_fn , mod_fn );
         if answer
            obj   = struct2cell( obj );    obj = {  obj{obj_id}};
            model = struct2cell(model);  model = {model{mod_id}};
            for i=1:length(obj)
               [ answer, msg ] = has_model( obj{i} , model{i} , obj_fn{i} );
               if ~answer
                  msg = sprintf(  '.%s%s'  ,obj_fn{i}, msg );
                  return
               end
            end
         else
            msg =  ' (struct fields don''t match the required fields)'  ;   
         end
      case  'cell'  
         answer   = isequal( size(obj), size(model) );
         if answer     
            for i=1:length(obj)
               [ answer, msg ] = has_model( obj{i} , model{i} ,  ''  );
               if ~answer
                  msg = sprintf(  '{%d}%s'  , i , msg );
                  return
               end
            end
         else
            msg =  ' (cell-array has not the required size)'  ;   
         end
      end
   end

   
   
function [ answer, msg ] = is_cell_of( obj , obj_type , obj_name ) 
   msg    =  ''  ;
   answer = isa( obj, obj_type );
   if ~answer
      answer = iscell( obj );
      if answer
         is_type = cellfun(  'isclass'  , obj ,  obj_type  );
         answer  = all( is_type );
         if ~answer
            bad_idx = find( ~is_type );
            msg     = sprintf(  '%s{%i}%s'  , obj_name , bad_idx(1) , msg );
         end
      end
   end
   
   
function [ answer, msg ] = is_recursive_cell_of( obj , obj_type , obj_name ) 
   msg    =  ''  ;
   answer = isa( obj, obj_type );
   if ~answer
      answer = iscell( obj );
      if answer
         is_type = cellfun(  'isclass'  , obj ,  obj_type  );
         is_cell = cellfun(  'isclass'  , obj ,  'cell'    );
         for i=find(is_cell)
            [ is_type(i), msg ] = is_recursive_cell_of( obj{i} , obj_type ,  ''  );
         end
         answer  = all( is_type );
         if ~answer
            bad_idx = find( ~is_type );
            msg     = sprintf(  '%s{%i}%s'  , obj_name , bad_idx(1) , msg );
         end
      end
   end


function answer = is_numeric( obj )
   answer = islogical( obj ) | isnumeric( obj );
